// Copyright © 2020 Birju Vachhani. All rights reserved.
// Use of this source code is governed by an Apache license that can be
// found in the LICENSE file.

import 'dart:async';
import 'dart:io';

import 'package:mockito/mockito.dart';
import 'package:path/path.dart' as p;
import 'package:spider/src/cli/models/spider_config.dart';
import 'package:spider/src/cli/process_terminator.dart';
import 'package:spider/src/cli/utils/utils.dart';
import 'package:spider/src/dart_class_generator.dart';
import 'package:test/test.dart';

import 'test_utils.dart';

void main() {
  final MockProcessTerminator processTerminatorMock = MockProcessTerminator();
  const Map<String, dynamic> testConfig = {
    "generate_tests": false,
    "no_comments": true,
    "export": true,
    "use_part_of": false,
    "use_references_list": false,
    "package": "resources",
    "groups": [
      {
        "class_name": "Assets",
        "sub_groups": [
          {
            "path": "assets/images",
            "types": ["png"],
            "prefix": "png"
          },
          {
            "path": "assets/images",
            "types": ["jpg"],
            "prefix": "jpg"
          }
        ]
      }
    ]
  };

  group('process tests', () {
    setUp(() {
      ProcessTerminator.setMock(processTerminatorMock);
    });

    test('asset generation test #1', () async {
      createTestConfigs(testConfig);
      createTestAssets();

      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator = DartClassGenerator(config.groups.first, config.globals);

      generator.process();
      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'assets.dart'));
      expect(genFile.existsSync(), isTrue);

      final classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Assets'));
      expect(classContent, contains('static const String pngTest1'));
      expect(classContent, contains('static const String jpgTest2'));
      expect(classContent, contains('assets/images/test1.png'));
      expect(classContent, contains('assets/images/test2.jpg'));
    });

    test('asset generation test - comments', () async {
      createTestConfigs(testConfig.copyWith({
        'no_comments': false,
      }));
      createTestAssets();

      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator = DartClassGenerator(config.groups.first, config.globals);

      generator.process();
      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'assets.dart'));
      expect(genFile.existsSync(), isTrue);

      final classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Assets'));
      expect(classContent, contains('static const String pngTest1'));
      expect(classContent, contains('static const String jpgTest2'));
      expect(classContent, contains('assets/images/test1.png'));
      expect(classContent, contains('assets/images/test2.jpg'));
      expect(classContent, contains('// Generated by spider'));
    });

    test('asset generation test no list of references', () async {
      createTestConfigs(testConfig.copyWith({"use_references_list": true}));
      createTestAssets();

      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator = DartClassGenerator(config.groups.first, config.globals);

      generator.process();
      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'assets.dart'));
      expect(genFile.existsSync(), isTrue);

      final classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Assets'));
      expect(classContent, contains('static const String pngTest1'));
      expect(classContent, contains('static const String jpgTest2'));
      expect(classContent, contains('assets/images/test1.png'));
      expect(classContent, contains('assets/images/test2.jpg'));
      expect(classContent, contains('static const List<String> values'));
    });

    test('asset generation test - watch', () async {
      createTestConfigs(testConfig.copyWith({
        'no_comments': false,
      }));
      createTestAssets();

      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator = DartClassGenerator(config.groups.first, config.globals);

      void proc() async {
        generator.initAndStart(watch: true, smartWatch: false);
      }

      proc();
      await Future.delayed(Duration(seconds: 5));

      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'assets.dart'));
      expect(genFile.existsSync(), isTrue);

      String classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Assets'));
      expect(classContent, contains('static const String pngTest1'));
      expect(classContent, contains('static const String jpgTest2'));
      expect(classContent, contains('assets/images/test1.png'));
      expect(classContent, contains('assets/images/test2.jpg'));
      expect(classContent, contains('// Generated by spider'));

      final newFile = File(p.join('assets', 'images', 'test3.png'));
      newFile.createSync();
      expect(newFile.existsSync(), isTrue);

      await Future.delayed(Duration(seconds: 5));
      classContent = genFile.readAsStringSync();

      expect(classContent, contains('static const String pngTest3'));
      expect(classContent, contains('assets/images/test3.png'));

      generator.cancelSubscriptions();
    });

    test('asset generation test - smart watch', () async {
      createTestConfigs(testConfig.copyWith({
        'no_comments': false,
        "groups": [
          {
            "class_name": "Assets",
            "sub_groups": [
              {
                "path": "assets/images",
                "types": ["jpg", "jpeg", "png", "webp", "gif", "bmp", "wbmp"]
              }
            ]
          },
          {
            "class_name": "Fonts",
            "sub_groups": [
              {
                "path": "assets/fonts",
              }
            ]
          }
        ]
      }));
      createTestAssets();
      createMoreTestAssets();

      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator1 =
          DartClassGenerator(config.groups.first, config.globals);

      final generator2 = DartClassGenerator(config.groups[1], config.globals);

      void proc1() async {
        generator1.initAndStart(watch: false, smartWatch: true);
      }

      void proc2() async {
        generator2.initAndStart(watch: false, smartWatch: true);
      }

      proc1();
      proc2();
      await Future.delayed(Duration(seconds: 5));

      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'fonts.dart'));
      expect(genFile.existsSync(), isTrue);

      String classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Fonts'));
      expect(classContent, contains('static const String test1'));
      expect(classContent, contains('static const String test2'));
      expect(classContent, contains('assets/fonts/test1.otf'));
      expect(classContent, contains('assets/fonts/test2.otf'));

      final String timestamp = genFile.readAsLinesSync().first;

      final newFile = File(p.join('assets', 'images', 'test3.png'));
      newFile.createSync();
      expect(newFile.existsSync(), isTrue);

      await Future.delayed(Duration(seconds: 5));
      classContent =
          File(p.join('lib', 'resources', 'assets.dart')).readAsStringSync();

      expect(classContent, contains('static const String test3'));
      expect(classContent, contains('assets/images/test3.png'));

      final String newTimestamp = genFile.readAsLinesSync().first;
      expect(timestamp, equals(newTimestamp));

      generator1.cancelSubscriptions();
      generator2.cancelSubscriptions();
    });

    test('asset generation test - test cases', () async {
      createTestConfigs(testConfig.copyWith({
        'generate_tests': true,
      }));

      createTestAssets();
      final Result<SpiderConfiguration> result = retrieveConfigs();
      expect(result.isSuccess, isTrue,
          reason: 'valid config file should not return error but it did.');

      final SpiderConfiguration config = result.data;

      final generator = DartClassGenerator(config.groups.first, config.globals);

      generator.process();
      verifyNever(processTerminatorMock.terminate(any, any));
      final genFile = File(p.join('lib', 'resources', 'assets.dart'));
      expect(genFile.existsSync(), isTrue);

      final classContent = genFile.readAsStringSync();

      expect(classContent, contains('class Assets'));
      expect(classContent, contains('static const String pngTest1'));
      expect(classContent, contains('static const String jpgTest2'));
      expect(classContent, contains('assets/images/test1.png'));
      expect(classContent, contains('assets/images/test2.jpg'));

      final genTestFile = File(p.join('test', 'assets_test.dart'));
      expect(genTestFile.existsSync(), isTrue);

      addTearDown(() => genTestFile.deleteSync());

      final testContent = genTestFile.readAsStringSync();

      expect(testContent, contains("import 'dart:io';"));
      expect(testContent, contains("import 'package:test/test.dart';"));
      expect(testContent,
          contains("import 'package:spider/resources/assets.dart';"));
      expect(testContent, contains("void main()"));
      expect(testContent,
          contains("expect(File(Assets.pngTest1).existsSync(), true);"));
      expect(testContent,
          contains("expect(File(Assets.jpgTest2).existsSync(), true);"));
    });

    tearDown(() {
      ProcessTerminator.clearMock();
      reset(processTerminatorMock);
      deleteTestAssets();
      deleteConfigFiles();
      deleteGeneratedRefs();
    });
  });
}
